'''This module defines Easy Equities data access components.'''
from datetime import datetime
from logging import debug, info
import json
import pandas as pd
from easy_equities_client.clients import EasyEquitiesClient
from easy_equities_client.accounts.types import Account
from frostaura.data_access.resources_data_access import IResourcesDataAccess
from frostaura.data_access.personal_asset_data_access import IPersonalAssetDataAccess

class EasyEquitiesPersonalAssetDataAccess(IPersonalAssetDataAccess):
    '''EasyEquities-related functionality.'''

    def __init__(self, resource_data_access: IResourcesDataAccess, username: str, password: str, config: dict = {}):
        self.resource_data_access = resource_data_access
        self.client = EasyEquitiesClient()
        self.username = username
        self.password = password
        self.config = config

    def __get_accounts__(self) -> list:
        info(f'Signing into EasyEquities as "{self.username}".')
        self.client.login(username=self.username, password=self.password)
        
        return self.client.accounts.list()

    def __get_account_by_name__(self, account_name: str) -> Account:
        accounts: list = self.__get_accounts__()
        
        debug(f'Filtering accounts list by name "{account_name}".')
        return [a for a in accounts if a.name == account_name][0]

    def __get_account_holdings__(self, account_name: str) -> list:
        account: Account = self.__get_account_by_name__(account_name=account_name)
        
        info(f'Getting account holdings for account "{account.name}" ({account.id}).')
        return self.client.accounts.holdings(account.id, include_shares=True)

    def __get_account_valuation__(self, account_name: str) -> list:
        account: Account = self.__get_account_by_name__(account_name=account_name)
        
        info(f'Getting account valuation for account "{account.name}" ({account.id}).')
        return self.client.accounts.valuations(account.id)

    def __get_excluded_assets__(self) -> list:
        with self.resource_data_access.get_resource(path='easy_equities_us_exclusions.json') as file:
            excluded_symbols: list = json.load(file)
            
            return excluded_symbols

    def get_supported_assets(self) -> pd.DataFrame:
        '''Get all supported asset names and symbols.'''

        info('Fetching EasyEquities supported symbols from a static source.')

        excluded_symbols: list = self.__get_excluded_assets__()

        with self.resource_data_access.get_resource(path='easy_equities_us_stocks.json') as file:
            company_names_symbols: list = json.load(file)
            data: dict = {
                'name': [c[0] for c in company_names_symbols if c[1] not in excluded_symbols],
                'symbol': [c[1] for c in company_names_symbols if c[1] not in excluded_symbols],
            }

            return pd.DataFrame(data)

    def get_personal_transactions(self) -> pd.DataFrame:
        '''Get all personal transactions made on an EasyEquities account.'''

        data: list = [
            ['Tesla Inc.', 'TSLA', 0.0688, datetime(2022, 7, 28, 0, 0)],
            ['Apple Inc.', 'AAPL', 0.4317, datetime(2022, 7, 1, 0, 0)],
            ['3D Systems Corporation', 'DDD', 8.8925, datetime(2022, 8, 5, 0, 0)],
            ['3D Systems Corporation', 'DDD', 5.2957, datetime(2022, 8, 18, 0, 0)],
            ['SPDR S&P 500 ETF', 'SPY', 0.0829, datetime(2022, 8, 22, 0, 0)],
            ['3D Systems Corporation', 'DDD', 2.317, datetime(2022, 8, 22, 0, 0)],
            ['Tesla Inc.', 'TSLA', 0.0884, datetime(2022, 8, 29, 0, 0)],
            ['SPDR S&P 500 ETF', 'SPY', 0.0547, datetime(2022, 8, 29, 0, 0)]
        ]

        return pd.DataFrame(data, columns =['name', 'symbol', 'shares', 'date']).set_index('date')



'''This module defines calculations engine components.'''
from frostaura.models import ProfitCalculationResult
import pandas as pd

class IAssetCalculationsEngine:
    '''Component to perform functions related to asset calculations.'''

    def calculate_holdings_profits(self, holdings: pd.DataFrame) -> pd.DataFrame:
        '''Determine individual asset profit ratio & profit USD and interpolate them into a copy of the given holdings.'''

        raise NotImplementedError()

    def calculate_holdings_profit(self, holdings: pd.DataFrame) -> ProfitCalculationResult:
        '''Determine the holdings profit percentage & profit USD, given the holdings.'''

        raise NotImplementedError()



'''This module defines the interface for the notifications data access components.'''
import matplotlib.pyplot as plt
import pandas as pd

class INotificationsDataAccess:
    '''Component to perform resource related actions.'''

    def send_figure(self, figure: plt.Figure):
        '''Send a Matplotlib figure as an image.'''

        raise NotImplementedError()

    def send_text(self, text: str) -> object:
        '''Send a text notification.'''

        raise NotImplementedError()

    def send_dataframe(self, dataframe: pd.DataFrame):
        '''Send a Pndas dataframe table.'''

        raise NotImplementedError()



'''This module defines profit calculation model components.'''

class ProfitCalculationResult:
    '''A model with all the post-profit calculation data for an asset.'''

    def __init__(self, percentage: float, value: float):
        self.percentage = percentage
        self.value = value



'''This module defines the interface for the notifications data access components.'''
from telegram import ParseMode
from telegram.ext import Updater
import io
import matplotlib.pyplot as plt
import prettytable as pt
import pandas as pd
from frostaura.data_access.notifications_data_access import INotificationsDataAccess

class TelegramNotificationsDataAccess(INotificationsDataAccess):
    '''Component to perform resource related actions.'''

    def __init__(self, bot_token: str, chat_id: str, register_handlers = None):
        assert bot_token is not None

        self.bot_token = bot_token
        self.register_handlers = register_handlers

        self.updater = Updater(token=self.bot_token)
        self.dispatcher = self.updater.dispatcher
        self.bot = self.updater.bot
        self.chat_id = chat_id

        if self.register_handlers is not None:
            self.register_handlers(self.dispatcher)

        self.updater.start_polling()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, t_b):
        self.updater.stop()
        self.updater = None
        self.dispatcher = None
        self.bot = None

    def __get_figure_buffer__(self, figure: plt.Figure) -> io.BytesIO:
        buffer = io.BytesIO()
        figure.savefig(buffer, format='png')
        buffer.seek(0)

        return buffer

    def send_text(self, text: str):
        '''Send a text notification.'''

        self.bot.send_message(chat_id=self.chat_id, text=text, parse_mode=ParseMode.HTML)

    def send_figure(self, figure: plt.Figure):
        self.bot.send_photo(chat_id=self.chat_id, photo=self.__get_figure_buffer__(figure))

    def send_dataframe(self, dataframe: pd.DataFrame):
        table = pt.PrettyTable(tuple(dataframe.columns))
        
        for index, row in dataframe.iterrows():
            row_data = list()

            for column_name in dataframe.columns:
                row_data.append(row[column_name])
            
            table.add_row(row_data)

        self.bot.send_message(chat_id=self.chat_id, text=f'<pre>{table}</pre>', parse_mode=ParseMode.HTML)



'''This module defines private manager components.'''
from cProfile import label
from datetime import datetime
import pandas as pd
from frostaura.data_access.notifications_data_access import INotificationsDataAccess
from frostaura.engines.visualization_engine import IVisualizationEngine
from frostaura.managers.asset_reporting_manager import IAssetReportingManager
from frostaura.data_access.personal_asset_data_access import IPersonalAssetDataAccess
from frostaura.engines.asset_calculations_engine import IAssetCalculationsEngine
from frostaura.engines.asset_projection_engine import IAssetProjectionEngine
from frostaura.models.profit_calculation_result import ProfitCalculationResult
from frostaura.engines.asset_valuation_engine import IAssetValuationEngine
from frostaura.data_access.public_asset_data_access import IPublicAssetDataAccess
from frostaura.models.symbol_data import SymbolData
from frostaura.models.visualization_types import VisualizationType
from frostaura.models.valuation_result import ValuationResult

class PersonalAssetReportingManager(IAssetReportingManager):
    '''Component to perform functions related to personal asset reporting.'''

    def __init__(self,
                 personal_asset_data_access: IPersonalAssetDataAccess,
                 asset_calculation_engine: IAssetCalculationsEngine,
                 asset_valuation_engine: IAssetValuationEngine,
                 asset_projection_engine: IAssetProjectionEngine,
                 visualization_engine: IVisualizationEngine,
                 personal_notification_data_access: INotificationsDataAccess,
                 public_asset_data_access: IPublicAssetDataAccess,
                 config: dict = {}):
        self.personal_asset_data_access = personal_asset_data_access
        self.asset_calculation_engine = asset_calculation_engine
        self.asset_valuation_engine = asset_valuation_engine
        self.asset_projection_engine = asset_projection_engine
        self.visualization_engine = visualization_engine
        self.personal_notification_data_access = personal_notification_data_access
        self.public_asset_data_access = public_asset_data_access
        self.config = config

    def __send_holdings_pie_chart__(self, holdings_with_profits: pd.DataFrame,):
        pie_fig, pie_ax = self.visualization_engine.get_figure(x='symbol',
                                                               y='ratio',
                                                               data=holdings_with_profits,
                                                               graph_type=VisualizationType.PIE,
                                                               title='Portfolio Assets')

        self.personal_notification_data_access.send_figure(figure=pie_fig)

    def __send_holdings_growth_projections__(self,
                                             holdings_with_profits: pd.DataFrame,
                                             n_months: int,
                                             monthly_deposits: list):
        holdings_growth = self.asset_projection_engine.project_monthly_holdings_growth(n_months=n_months,
                                                                                       holdings_with_profits=holdings_with_profits,
                                                                                       monthly_deposits=monthly_deposits)

        currency_format: str = '${x:,.0f}'
        fig, ax = self.visualization_engine.get_figure(x='month',
                                                       y='balance',
                                                       data=holdings_growth,
                                                       graph_type=VisualizationType.LINE,
                                                       title=f'Holdings Growth Projection | {int(n_months/12)} Years',
                                                       legend=True,
                                                       line_label=f'Balance | {currency_format.format(x=holdings_growth.iloc[-1]["balance"])}',
                                                       y_tick_format_str=currency_format)

        for i, r in holdings_growth.iterrows():
            # Draw a line for every 5 year.
            if r['month'] % (12 * 5) == 0:
                year: int = int(r['month'] / 12)
                ax.axvline(x=r['month'], linestyle='--', color='#11E5AD', label=f'Year {year} | {currency_format.format(x=r["balance"])}')

        fig.legends = [fig.legend()]
        percentage_increase_to_projected_time: float = round((holdings_growth.iloc[-1]['balance'] - holdings_growth.iloc[0]['balance']) / holdings_growth.iloc[0]['balance'] * 100, 2)

        self.personal_notification_data_access.send_text(f'<strong>Holdings Growth Projection ({int(n_months/12)} Years) |</strong> {percentage_increase_to_projected_time}%')
        self.personal_notification_data_access.send_figure(figure=fig)

    def __send_holdings_report__(self, holdings: list):
        overall_holdings_profit: ProfitCalculationResult = self.asset_calculation_engine.calculate_holdings_profit(holdings)

        self.personal_notification_data_access.send_text(f'<strong>Overall Holdings Profits:</strong> ${round(overall_holdings_profit.value, 2)} ({round(overall_holdings_profit.percentage if overall_holdings_profit.value > 0 else -overall_holdings_profit.percentage)}%)')

    def __send_wealth_projections_report__(self,
                                         holdings_with_profits: pd.DataFrame,
                                         monthly_deposits: float):
        wealth_projection = self.asset_projection_engine.project_wealth_growth(holdings_with_profits=holdings_with_profits,
                                                                               monthly_deposits=monthly_deposits)

        wealth_projection['Target Balance (USD)'] = wealth_projection['Target Balance (USD)'].apply(lambda v: int(v))
        wealth_projection['ETA (Years)'] = wealth_projection['ETA (Years)'].apply(lambda v: int(v))
        currency_format: str = '${x:,.0f}'
        fig, ax = self.visualization_engine.get_figure(x='ETA (Years)',
                                                       y='Target Balance (USD)',
                                                       data=wealth_projection,
                                                       graph_type=VisualizationType.LINE,
                                                       title='Wealth Projection | 30 Years',
                                                       legend=True,
                                                       line_label=wealth_projection.iloc[-1]['Target Balance (USD)'],
                                                       y_tick_format_str=currency_format)

        for i, r in wealth_projection.iterrows():
            ax.axhline(y=r['Target Balance (USD)'], linestyle='-', color='#11E5AD', label=currency_format.format(x=r['Target Balance (USD)']))

        fig.legends = [fig.legend()]

        self.personal_notification_data_access.send_text(text=f'<strong>Wealth Projection |</strong> 30 Years')
        self.personal_notification_data_access.send_figure(figure=fig)
        self.personal_notification_data_access.send_dataframe(dataframe=wealth_projection[['Target Balance (USD)', 'ETA (Months)', 'ETA (Years)']]
                                                                .rename(columns={
                                                                    'Target Balance (USD)': 'USD',
                                                                    'ETA (Months)': 'Months',
                                                                    'ETA (Years)': 'Years'
                                                                }))

    def __send_individual_asset_performance_reports__(self,
                                                      holdings: list,
                                                      holdings_with_profits: pd.DataFrame):
        currency_format: str = '${x:,.0f}'
        green: str = '#11E5AD'
        red: str = '#e55111'

        for row_index, row in holdings_with_profits.iterrows():
            symbol: str = row['symbol']
            company: str = row['name']
            current_total: float = currency_format.format(x=row['total_current_usd'])
            profit_percentage: float = round(row['total_profit_ratio'], 2)
            profit_absolute: float = row['total_profit_usd']
            transactions_for_symbol: list = holdings.loc[holdings['symbol'] == symbol]
            first_transaction_date: datetime = transactions_for_symbol.iloc[0].name
            history: pd.DataFrame = self.public_asset_data_access.get_symbol_history(symbol=symbol)
            history = history[first_transaction_date:].reset_index()

            if profit_absolute < 0:
                profit_percentage = -profit_percentage

            profit_absolute = currency_format.format(x=profit_absolute)

            message: str = f'<strong>{company} ({symbol})</strong> holdings currently amount to {current_total}. A {profit_absolute} ({profit_percentage}%) profit.'

            self.personal_notification_data_access.send_text(text=message)

            symbol_data: SymbolData = self.public_asset_data_access.get_symbol_info(symbol=symbol)
            symbol_valuation: ValuationResult = self.asset_valuation_engine.valuate(symbol_data=symbol_data)

            if symbol_valuation is None:
                return

            fig, ax = self.visualization_engine.get_figure(x='Date',
                                                           y='Close',
                                                           data=history,
                                                           graph_type=VisualizationType.LINE,
                                                           title=f'{company} ({symbol})',
                                                           subtitle=None if (symbol_valuation.annual_dividend_percentage is None or symbol_valuation.divident_payout_frequency_in_months == 0) else f'Dividend: {round(symbol_valuation.annual_dividend_percentage, 2)}% Annually ({symbol_valuation.divident_payout_frequency_in_months} Month Frequency)',
                                                           legend=True,
                                                           line_label=f'Current Value: {currency_format.format(x=history.iloc[-1]["Close"])} (Δ{round(profit_percentage, 2)}%)',
                                                           y_tick_format_str=currency_format)

            # Draw valuation line
            ax.axhline(y=symbol_valuation.fair_price,
                       linestyle='-',
                       color=red if symbol_valuation.is_overvalued else green,
                       label=f'Intrinsic Value: {currency_format.format(x=symbol_valuation.fair_price)} (Δ{round(symbol_valuation.absolute_current_v_valuation_delta*100, 2)}%)')

            for trow_index, trow in transactions_for_symbol.iterrows():
                transaction_date: datetime = trow.name
                price_at_transaction_date: float = history.loc[history['Date'] == transaction_date]['Close'].values[0]

                ax.plot(transaction_date,
                        price_at_transaction_date,
                        'ro',
                        color=red if price_at_transaction_date > symbol_valuation.current_price else green)

            fig.legends = [fig.legend()]

            self.personal_notification_data_access.send_figure(figure=fig)

    def send_reports(self):
        '''Generate and send asset reports.'''

        holdings: list = self.personal_asset_data_access.get_personal_transactions()
        holdings_with_profits: pd.DataFrame = self.asset_calculation_engine.calculate_holdings_profits(holdings=holdings)
        n_months_to_project: int = 12 * 5 # 5 Years
        usd_zar_exchange_rate: float = 16.17
        monthly_zar_deposit: float = 2000
        monthly_deposits: float = [(monthly_zar_deposit / usd_zar_exchange_rate / len(holdings_with_profits)) for h in holdings_with_profits]

        self.__send_holdings_report__(holdings=holdings)
        self.__send_holdings_pie_chart__(holdings_with_profits=holdings_with_profits)
        self.__send_holdings_growth_projections__(n_months=n_months_to_project,
                                                  monthly_deposits=monthly_deposits,
                                                  holdings_with_profits=holdings_with_profits)
        self.__send_individual_asset_performance_reports__(holdings=holdings,
                                                           holdings_with_profits=holdings_with_profits)
        self.__send_wealth_projections_report__(holdings_with_profits=holdings_with_profits,
                                                monthly_deposits=monthly_deposits)



'''This module defines Yahoo Finance data access components.'''
from frostaura.models.symbol_data import SymbolData
import yfinance as yf
import pandas as pd
import time
import requests
import json
from joblib import Parallel, delayed
from logging import info, warning
from frostaura.data_access.public_asset_data_access import IPublicAssetDataAccess

class YahooFinanceDataAccess(IPublicAssetDataAccess):
    '''Yahoo Finance public asset-related functionality.'''

    def __init__(self, config: dict = {}):
        self.config = config

    def __get_value__(self, root: dict, path: str, default: object = None) -> object:
        segments: list = path.split('.')
        context: dict = root

        for segment in segments:
            if not segment in context:
                return default

            context = context[segment]
        
        return context

    def __get_symbol_quote__(self, symbol: str, current_attempt: int=1, max_retry_count: int=3) -> dict:
        try:
            base_url: str = f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/{symbol}?modules=assetProfile,defaultKeyStatistics,financialData,earningsTrend,price,summaryDetail'
            http_response = requests.get(base_url, headers={
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
            })
            data = json.loads(http_response.text)
            symbol_quote = data['quoteSummary']['result'][0]
            response: dict = {
                'symbol': symbol,
                'eps': self.__get_value__(root=symbol_quote, path='defaultKeyStatistics.trailingEps.raw'),
                'long_description': self.__get_value__(root=symbol_quote, path='assetProfile.longBusinessSummary'),
                'company_name': self.__get_value__(root=symbol_quote, path='price.longName'),
                'market_cap': self.__get_value__(root=symbol_quote, path='price.marketCap.raw'),
                'current_price': self.__get_value__(root=symbol_quote, path='financialData.currentPrice.raw'),
                'total_cash': self.__get_value__(root=symbol_quote, path='financialData.totalCash.raw'),
                'total_debt': self.__get_value__(root=symbol_quote, path='financialData.totalDebt.raw'),
                'total_revenue': self.__get_value__(root=symbol_quote, path='financialData.totalRevenue.raw'),
                'pe_ratio': self.__get_value__(root=symbol_quote, path='summaryDetail.trailingPE.raw'),
                'ex_dividend_date': self.__get_value__(root=symbol_quote, path='summaryDetail.exDividendDate.raw'),
                'dividend_yield': self.__get_value__(root=symbol_quote, path='summaryDetail.dividendYield.raw'),
                'growth_rate': None
            }

            five_year_growth_periods: list = [t for t in self.__get_value__(root=data['quoteSummary']['result'][0], path='earningsTrend.trend', default=[]) if t['period'] == '+5y']

            if len(five_year_growth_periods) > 0:
                response['growth_rate'] = self.__get_value__(root=five_year_growth_periods[0], path='growth.raw')

            return response
        except Exception as ex:
            if current_attempt <= max_retry_count:
                print(f'Symbol "{symbol}" failed: "{ex}". Retrying {current_attempt}/{max_retry_count}.')
                time.sleep(current_attempt)

                return self.__get_symbol_quote__(symbol=symbol, current_attempt=current_attempt+1)

            return None

    def __get_symbol_quotes_async__(self, symbols: list) -> list:
        parrallelizer: Parallel = Parallel(n_jobs=len(symbols), prefer='threads')
        results: list = parrallelizer([delayed(self.__get_symbol_quote__)(s) for s in symbols])

        return results

    def get_symbol_history(self, symbol: str) -> pd.DataFrame:
        '''Get historical price movements for a given symbol.'''

        info(f'Fetching historical price movements for symbol "{symbol}".')

        ticker = yf.Ticker(symbol)
        value = ticker.history(period='max')

        return value

    def get_symbol_info(self, symbol: str) -> SymbolData:
        '''Get the data for a specific symbol.'''
        quote: dict = self.__get_symbol_quote__(symbol=symbol)
            
        return SymbolData(symbol=quote['symbol'],
                          company_name=quote['company_name'],
                          current_price=quote['current_price'],
                          eps=quote['eps'],
                          annual_growth_projected=quote['growth_rate'])

    def augment_symbols_info(self, symbols: pd.DataFrame) -> pd.DataFrame:
        '''Add info to a daraframe containing symbols.'''
        def get_symbol_data(symbol: str) -> SymbolData:
            quote: dict = self.__get_symbol_quote__(symbol=symbol)
            
            return SymbolData(symbol=quote['symbol'],
                              company_name=quote['company_name'],
                              current_price=quote['current_price'],
                              eps=quote['eps'],
                              annual_growth_projected=quote['growth_rate'])
        
        __symbols__: pd.DataFrame = symbols.copy()
        __symbols__['quote'] = __symbols__['symbol'].apply(get_symbol_data)

        return __symbols__.sort_values('symbol').loc[__symbols__['quote'] != None]



'''This module defines HTTP resources data access component.'''
import requests
from bs4 import BeautifulSoup
from logging import debug, info
from frostaura.data_access.resources_data_access import IResourcesDataAccess

class HtmlResourcesDataAccess(IResourcesDataAccess):
    '''Component to perform HTML resource related actions.'''

    def __init__(self, config: dict = {}):
        self.config = config
        self.default_user_agent = 'PostmanRuntime/7.29.0'

    def get_resource(self, path: str) -> BeautifulSoup:
        '''Get a queryable HTML page via a URL (path).'''

        info(f'Fetching HTML page from URL "{path}".')

        user_agent_setting_key: str = 'user-agent'

        if user_agent_setting_key not in self.config:
            debug(f'No key "{user_agent_setting_key}" found in config. Defaulting to user agent value "{self.default_user_agent}".')

            self.config[user_agent_setting_key] = self.default_user_agent
        else:
            debug(f'User agent override found in config. Using value "{self.config[user_agent_setting_key]}".')

        response: requests.Response = requests.get(url=path, headers={
            'User-Agent': self.config[user_agent_setting_key]
        })
        response_text: str = response.text

        return BeautifulSoup(response_text, 'html.parser')



'''This module defines public manager components.'''
import pandas as pd
from frostaura.data_access.notifications_data_access import INotificationsDataAccess
from frostaura.engines.visualization_engine import IVisualizationEngine
from frostaura.managers.asset_reporting_manager import IAssetReportingManager
from frostaura.data_access.personal_asset_data_access import IPersonalAssetDataAccess
from frostaura.engines.asset_valuation_engine import IAssetValuationEngine
from frostaura.data_access.public_asset_data_access import IPublicAssetDataAccess
from frostaura.models.visualization_types import VisualizationType
from frostaura.models.valuation_result import ValuationResult

class PublicAssetReportingManager(IAssetReportingManager):
    '''Component to perform functions related to public asset reporting.'''

    def __init__(self,
                 personal_asset_data_access: IPersonalAssetDataAccess,
                 asset_valuation_engine: IAssetValuationEngine,
                 visualization_engine: IVisualizationEngine,
                 public_notification_data_access: INotificationsDataAccess,
                 public_asset_data_access: IPublicAssetDataAccess,
                 config: dict = {}):
        self.personal_asset_data_access = personal_asset_data_access
        self.asset_valuation_engine = asset_valuation_engine
        self.visualization_engine = visualization_engine
        self.public_notification_data_access = public_notification_data_access
        self.public_asset_data_access = public_asset_data_access
        self.config = config

    def __send_individual_asset_performance_reports__(self, symbol_data: list):
        currency_format: str = '${x:,.0f}'
        green: str = '#11E5AD'
        red: str = '#e55111'
        top: int = 15
        top_x_symbol_data = symbol_data.copy()
        top_x_symbol_data = [s for s in top_x_symbol_data if s['valuation'] is not None and not s['valuation'].is_overvalued]

        if len(top_x_symbol_data) <= 0:
            self.public_notification_data_access.send_text(text='No undervalued stocks to report currently.')
            return

        for data in top_x_symbol_data:
            valuation: ValuationResult = data['valuation']

            if valuation.is_overvalued:
                valuation.absolute_current_v_valuation_delta = -valuation.absolute_current_v_valuation_delta

        top_x_symbol_data = sorted(top_x_symbol_data,
                             key=lambda i: i['valuation'].absolute_current_v_valuation_delta,
                             reverse=True)[:top]

        text: str = f'Top <i>{top}</i> EasyEquities <strong>Value</strong> Assets <i>(better viewed horizontally)</i>'
        self.public_notification_data_access.send_text(text=text)
        assets_table: dict = {
            'Symbol': [],
            'Close': [],
            'Intrinsic': [],
            '%': [],
            'Dividend': []
        }

        for data in top_x_symbol_data:
            valuation: ValuationResult = data['valuation']
            symbol: str = valuation.symbol
            company: str = valuation.company_name

            assets_table['Symbol'].append(symbol)
            assets_table['Close'].append(currency_format.format(x=valuation.current_price))
            assets_table['Intrinsic'].append(currency_format.format(x=valuation.fair_price))
            assets_table['%'].append(f'{round(valuation.absolute_current_v_valuation_delta * 100, 2)}%')

            if valuation.divident_payout_frequency_in_months > 0:
                assets_table['Dividend'].append(f'{round(valuation.annual_dividend_percentage, 2)}%')
            else:
                assets_table['Dividend'].append(currency_format.format(x=0))

        self.public_notification_data_access.send_dataframe(pd.DataFrame(assets_table))

        for data in top_x_symbol_data:
            valuation: ValuationResult = data['valuation']
            symbol: str = valuation.symbol
            company: str = valuation.company_name
            history: pd.DataFrame = self.public_asset_data_access.get_symbol_history(symbol=symbol)

            fig, ax = self.visualization_engine.get_figure(x='Date',
                                                           y='Close',
                                                           data=history,
                                                           graph_type=VisualizationType.LINE,
                                                           title=f'{company} ({symbol})',
                                                           subtitle=None if (valuation.annual_dividend_percentage is None or valuation.divident_payout_frequency_in_months == 0) else f'Dividend: {round(valuation.annual_dividend_percentage, 2)}% Annually ({valuation.divident_payout_frequency_in_months} Month Frequency)',
                                                           legend=True,
                                                           line_label=f'Current Value: {currency_format.format(x=valuation.current_price)}',
                                                           y_tick_format_str=currency_format)

            # Draw valuation line
            ax.axhline(y=valuation.fair_price,
                       linestyle='-',
                       color=red if valuation.is_overvalued else green,
                       label=f'Intrinsic Value: {currency_format.format(x=valuation.fair_price)} (Δ{round(valuation.absolute_current_v_valuation_delta*100, 2)}%)')

            fig.legends = [fig.legend()]

            self.public_notification_data_access.send_figure(figure=fig)

    def send_reports(self):
        '''Generate and send asset reports.'''

        all_symbols: list = self.personal_asset_data_access.get_supported_assets()
        symbols_with_info: pd.DataFrame = self.public_asset_data_access.augment_symbols_info(symbols=all_symbols)
        symbol_data: list = [{ 'history': None, 'valuation': self.asset_valuation_engine.valuate(symbol_data=v) } for v in list(symbols_with_info.quote.values)]
        symbol_data = [sd for sd in symbol_data if sd['valuation'] is not None]

        self.__send_individual_asset_performance_reports__(symbol_data=symbol_data)



'''This module defines valuation engine components.'''
from frostaura.data_access import IPublicAssetDataAccess
from frostaura.data_access import IResourcesDataAccess
from frostaura.models.valuation_result import ValuationResult
from frostaura.models.symbol_data import SymbolData
from frostaura.engines.asset_valuation_engine import IAssetValuationEngine

class GrahamValuationEngine(IAssetValuationEngine):
    '''Valuation-related functionality using the discounted free cash flow method.'''
    margin_of_safety: float = 0.5

    def __init__(self, html_data_access: IResourcesDataAccess, public_asset_data_access: IPublicAssetDataAccess, config: dict = {}):
        self.html_data_access = html_data_access
        self.public_asset_data_access = public_asset_data_access
        self.config = config

    def __determine_intrinsic_value__(self,
                                      eps: float,
                                      pe_base_non_growth_company: float,
                                      annual_growth_projected: float,
                                      average_yield_of_aaa_corporate_bonds: float,
                                      current_yield_of_aaa_corporate_bonds: float) -> float:
        value: float = eps * (pe_base_non_growth_company + 2 * annual_growth_projected) * average_yield_of_aaa_corporate_bonds / current_yield_of_aaa_corporate_bonds

        return value

    def valuate(self, symbol_data: SymbolData) -> ValuationResult:
        '''Valuate a given asset.'''
        try:
            valuation: float = self.__determine_intrinsic_value__(eps=symbol_data.eps,
                                                                pe_base_non_growth_company=symbol_data.pe_base_non_growth_company,
                                                                annual_growth_projected=symbol_data.annual_growth_projected*100,
                                                                average_yield_of_aaa_corporate_bonds=symbol_data.average_yield_of_aaa_corporate_bonds,
                                                                current_yield_of_aaa_corporate_bonds=symbol_data.current_yield_of_aaa_corporate_bonds)

            return ValuationResult(
                symbol=symbol_data.symbol,
                company_name=symbol_data.company_name,
                current_price=symbol_data.current_price,
                valuation_price=valuation,
                valuation_method='benjamin_graham_valuation',
                margin_of_safety=self.margin_of_safety
            )
        except Exception:
            return None



'''This module defines personal asset data access components.'''
import pandas as pd

class IPersonalAssetDataAccess:
    '''Component to perform functions related to personal / owned assets.'''

    def get_supported_assets(self) -> pd.DataFrame:
        '''Get all supported asset names and symbols.'''

        raise NotImplementedError()

    def get_personal_transactions(self) -> pd.DataFrame:
        '''Get all personal transactions made on an EasyEquities account.'''

        raise NotImplementedError()



'''A convenience importing mechanism for all manager components.'''

from .asset_reporting_manager import IAssetReportingManager
from .asset_reporting_manager__personal import PersonalAssetReportingManager
from .asset_reporting_manager__public import PublicAssetReportingManager



'''This module defines projection engine components.'''
import pandas as pd

class IAssetProjectionEngine:
    '''Component to perform functions related to asset projections.'''

    def project_monthly_asset_growth(self,
                                     n_months: int,
                                     symbol: str,
                                     principal_value: float,
                                     monthly_deposit: float) -> pd.DataFrame:
        '''Determine an asset's growth at a given annual rate over a specified number of months while applying a monthly deposit.'''

        raise NotImplementedError()

    def project_monthly_holdings_growth(self,
                                        n_months: int,
                                        holdings_with_profits: pd.DataFrame,
                                        monthly_deposits: list) -> pd.DataFrame:
        '''Determine a comprehensive holdings growth at a given annual rates over a specified numbers of months while applying a monthly deposits.'''

        raise NotImplementedError()

    def project_wealth_growth(self,
                              holdings_with_profits: pd.DataFrame,
                              monthly_deposits: list,
                              n_months: int,
                              min_amount: float,
                              max_amount: float,
                              factorials_of: int) -> pd.DataFrame:
        '''Determine wealth growth at a given annual rates over a specified numbers of months while applying a monthly deposits.'''

        raise NotImplementedError()



'''This module defines valuation engine components.'''
from frostaura.models.symbol_data import SymbolData
from frostaura.models.valuation_result import ValuationResult

class IAssetValuationEngine:
    '''Component to perform functions related to asset valuation.'''

    def valuate(self, symbol_data: SymbolData) -> ValuationResult:
        '''Valuate a given asset.'''

        raise NotImplementedError()



'''This module defines projection engine components.'''
import pandas as pd
from frostaura.models.symbol_data import SymbolData
import inflect
from frostaura.engines.asset_projection_engine import IAssetProjectionEngine
from frostaura.engines.asset_valuation_engine import IAssetValuationEngine
from frostaura.models.valuation_result import ValuationResult
from frostaura.data_access.public_asset_data_access import IPublicAssetDataAccess

class SimpleAssetProjectionEngine(IAssetProjectionEngine):
    '''Calculations-related functionality using some maths under-the-hood.'''

    def __init__(self,
                 asset_valuation_engine: IAssetValuationEngine,
                 public_asset_data_access: IPublicAssetDataAccess,
                 config: dict = {}):
        self.asset_valuation_engine = asset_valuation_engine
        self.public_asset_data_access = public_asset_data_access
        self.config = config
        self.inflect_engine = inflect.engine()

    def project_monthly_asset_growth(self,
                                     n_months: int,
                                     symbol: str,
                                     principal_value: float,
                                     monthly_deposit: float = 0) -> pd.DataFrame:
        '''Determine an asset's growth at a given annual rate over a specified number of months while applying a monthly deposit.'''

        symbol_data: SymbolData = self.public_asset_data_access.get_symbol_info(symbol=symbol)
        valuation: ValuationResult = self.asset_valuation_engine.valuate(symbol_data=symbol_data)
        annual_growth_rate: float = symbol_data.annual_growth_projected
        divident_payout_frequency_in_months: int = valuation.divident_payout_frequency_in_months
        data = {
            'month': list(),
            'deposits_withdrawals': list(),
            'interest': list(),
            'total_deposits_withdrawals': list(),
            'accrued_interest': list(),
            'balance': list()
        }

        data['month'].append(0)
        data['deposits_withdrawals'].append(0)
        data['interest'].append(0)
        data['total_deposits_withdrawals'].append(0)
        data['accrued_interest'].append(0)
        data['balance'].append(principal_value)

        for i in range(1, n_months + 1):
            data['month'].append(i)
            data['deposits_withdrawals'].append(monthly_deposit)
            
            interest: float = (annual_growth_rate / 12) * data['balance'][-1]

            if divident_payout_frequency_in_months > 0 and i % divident_payout_frequency_in_months == 0:
                annual_dividend_rate: float = valuation.annual_dividend_percentage / 100
                interest += (annual_dividend_rate / 12 / divident_payout_frequency_in_months) * data['balance'][-1]

            data['interest'].append(interest)
            data['accrued_interest'].append(data['accrued_interest'][-1] + data['interest'][-1])
            data['total_deposits_withdrawals'].append(data['total_deposits_withdrawals'][-1] + monthly_deposit)
            data['balance'].append(data['interest'][-1] + data['balance'][-1] + monthly_deposit)

        response: pd.DataFrame = pd.DataFrame(data)

        return response.set_index('month')

    def project_monthly_holdings_growth(self,
                                   n_months: int,
                                   holdings_with_profits: pd.DataFrame,
                                   monthly_deposits: list) -> pd.DataFrame:
        '''Determine a comprehensive holdings growth at a given annual rates over a specified numbers of months while applying a monthly deposits.'''

        annual_growth_rates: list = list()
        principal_values: list = list()
        symbols: list = list()

        for row_index, row in holdings_with_profits.iterrows():
            try:
                symbol_data: SymbolData = self.public_asset_data_access.get_symbol_info(symbol=row['symbol'])
                symbol_valuation: ValuationResult = self.asset_valuation_engine.valuate(symbol_data=symbol_data)
                annual_growth_rate: float = symbol_data.annual_growth_projected

                if symbol_valuation.annual_dividend_percentage is not None:
                    annual_growth_rate += (symbol_valuation.annual_dividend_percentage / 100)

                annual_growth_rates.append(annual_growth_rate)
                principal_values.append(row['total_current_usd'])
                symbols.append(row['symbol'])
            except Exception:
                pass

        projections: list = list()

        for month_index in range(len(annual_growth_rates)):
            projection = self.project_monthly_asset_growth(n_months=n_months,
                                                           symbol=symbols[month_index],
                                                           principal_value=principal_values[month_index],
                                                           monthly_deposit=monthly_deposits[month_index])

            projections.append(projection)

        data = {
            'month': list(),
            'deposits_withdrawals': list(),
            'interest': list(),
            'total_deposits_withdrawals': list(),
            'accrued_interest': list(),
            'balance': list()
        }

        for month_index in range(1, n_months + 1):
            data['month'].append(month_index)
            data['deposits_withdrawals'].append(sum([p.loc[month_index]['deposits_withdrawals'] for p in projections]))

            previous_row_balances: float = sum([p.loc[month_index - 1]['balance'] for p in projections])
            current_row_balances: float = sum([p.loc[month_index]['balance'] for p in projections])
            current_deposits: float = sum([p.loc[month_index]['deposits_withdrawals'] for p in projections])
            interest = 1 - previous_row_balances / (current_row_balances - current_deposits)

            data['interest'].append(interest)
            data['total_deposits_withdrawals'].append(sum([p.loc[month_index]['total_deposits_withdrawals'] for p in projections]))
            data['accrued_interest'].append(sum([p.loc[month_index]['accrued_interest'] for p in projections]))
            data['balance'].append(sum([p.loc[month_index]['balance'] for p in projections]))

        return pd.DataFrame(data)

    def project_wealth_growth(self,
                              holdings_with_profits: pd.DataFrame,
                              monthly_deposits: list,
                              n_months: int = 12 * 50, # 50 Years
                              min_amount: float = 1,
                              max_amount: float = 100 * 1000 * 10 * 1000 * 10, # Ten Billion
                              factorials_of: int = 10) -> pd.DataFrame:
        '''Determine wealth growth at a given annual rates over a specified numbers of months while applying a monthly deposits.'''

        factorials_of: int = 10
        holdings_growth: pd.DataFrame = self.project_monthly_holdings_growth(n_months=n_months,
                                                                             holdings_with_profits=holdings_with_profits,
                                                                             monthly_deposits=monthly_deposits)
        current_amount_to_project_calculation: float = min_amount
        data = {
            'Target Balance (USD)': list(),
            'Target Balance Eng (USD)': list(),
            'ETA (Months)': list(),
            'ETA (Years)': list()
        }
        current_balance: float = holdings_with_profits['total_current_usd'].sum()

        while current_amount_to_project_calculation <= max_amount:
            projections: int = holdings_growth.loc[holdings_growth['balance'] > current_amount_to_project_calculation]

            if current_balance > current_amount_to_project_calculation:
                current_amount_to_project_calculation *= factorials_of
                continue

            if projections.shape[0] <= 0:
                break

            data['Target Balance (USD)'].append(projections.iloc[0]["balance"])
            data['Target Balance Eng (USD)'].append(self.inflect_engine.number_to_words(projections.iloc[0]["balance"]))
            data['ETA (Months)'].append(projections.index[0])
            data['ETA (Years)'].append(projections.index[0] / 12)
            
            current_amount_to_project_calculation *= factorials_of

        return pd.DataFrame(data)


'''A convenience importing mechanism for all data access components.'''
from frostaura.data_access.resources_data_access import IResourcesDataAccess
from frostaura.data_access.resources_data_access__html import HtmlResourcesDataAccess
from frostaura.data_access.resources_data_access__embedded import EmbeddedResourcesDataAccess

from frostaura.data_access.public_asset_data_access import IPublicAssetDataAccess
from frostaura.data_access.public_asset_data_access__yahoo_finance import YahooFinanceDataAccess

from frostaura.data_access.personal_asset_data_access import IPersonalAssetDataAccess
from frostaura.data_access.personal_asset_data_access__easy_equities import EasyEquitiesPersonalAssetDataAccess

from frostaura.data_access.notifications_data_access import INotificationsDataAccess
from frostaura.data_access.notifications_data_access__telegram import TelegramNotificationsDataAccess



'''This module defines the interface for the resources data access components.'''

class IResourcesDataAccess:
    '''Component to perform resource related actions.'''

    def get_resource(self, path: str) -> object:
        '''Get a resource's content.'''

        raise NotImplementedError()



'''A convenience importing mechanism for all engine components.'''

from frostaura.engines.asset_valuation_engine import IAssetValuationEngine
from frostaura.engines.asset_calculations_engine import IAssetCalculationsEngine
from frostaura.engines.asset_calculations_engine__simple import SimpleAssetCalculationsEngine
from frostaura.engines.asset_projection_engine import IAssetProjectionEngine
from frostaura.engines.asset_projection_engine__simple import SimpleAssetProjectionEngine
from frostaura.engines.visualization_engine import IVisualizationEngine
from frostaura.engines.visualization_engine__frostaura import FrostAuraVisualizationEngine
from frostaura.engines.asset_valuation_engine__graham import GrahamValuationEngine
